# Function to read image from disk.
import cv2 as cv
import numpy as np


def imread(fp: str, grayscale: bool = False) -> np.ndarray:
    """Read image from disk. Currently does not support reading RGBA images.

    Args:
        fp (str): The file path to save the image.
        grayscale (bool, optional): If True, the image will be read as grayscale.
            Defaults to False.

    Returns:
        np.ndarray: The image read from disk.

    """
    if grayscale:
        img = cv.imread(fp, cv.IMREAD_GRAYSCALE)
    else:
        img = cv.imread(fp, cv.IMREAD_UNCHANGED)

        # Convert the image from OpenCV's BGR to RGB.
        shape = img.shape

        if len(shape) == 3:
            if shape[2] == 3:
                img = cv.cvtColor(img, cv.COLOR_BGR2RGB)
            elif shape[2] == 4:
                img = cv.cvtColor(img, cv.COLOR_RGBA2BGRA)

    return img



"""Functions for evaluating datasets with machine learning models.

- evaluate_semantic_segmentation_segformer_group: Evaluate a segformer
  trainer model on a specific dataset provided as a dataframe, grouped
  by a specific column.
- per_class_dice_on_dataset: Per class DICE coefficient on a set of
    images.

"""

import torch
from torch import nn
from tqdm import tqdm
from transformers.trainer import Trainer
import pandas as pd
import numpy as np
from .datasets.utils import create_segformer_segmentation_dataset
from .transforms.segformer_transforms import val_transforms
from ..tile_utils import merge_tile_masks


def evaluate_semantic_segmentation_segformer_group(
    trainer: Trainer,
    df: pd.DataFrame,
    group_col: str,
    batch_size: int = 16,
    tile_size: int = 512,
    background_label: int | None = None,
) -> float:
    """Evaluate a segformer trainer model on a specific dataset provided
    as a dataframe, grouped by a specific column. For a unique value of
    the column it will predict on the images for that group and then
    merge the predictions into a single multipolygon. The mean IoU will
    be calculated agains the ground truth masks for that group. The
    reported mean IoU will be the average of all the mean IoUs for each
    group.

    Args:
        trainer (Trainer): The trainer object with the model to evaluate.
        df (pandas.DataFrame): The dataframe with the tile metadata.
            Must have the fp column (file path), x column (x coordinate),
            y column (y coordinate), and group_col.
        group_col (str): The column to group the tiles by.
        batch_size (int): The batch size to use for prediction. Note that
            this is not the batch size use for predictions, rather it is
            how many masks will be stored in memory. Batch size used
            for predictions is specified on the trainer input. Default is
            16.
        tile_size (int): The size of the tiles. Default is 512.
        background_label (int | None): The label to use for the background
            class, which will be ignored. If None then no label will be
            ignored. Default is None.

    Returns:
        float: The mean IoU across all groups.

    """
    ious = []

    # Predict on the images for each group separately.
    groups = list(df[group_col].unique())
    for group in tqdm(groups, desc="Evaluating groups"):
        # Get the tiles in this group.
        group_df = df[df[group_col] == group]

        pred_tile_list = []
        true_tile_list = []

        # Predict on the group in batches.
        for i in range(0, len(group_df), batch_size):
            # Create dataset for this batch.
            batch_df = group_df.iloc[i : i + batch_size]
            x_list = batch_df["x"].tolist()
            y_list = batch_df["y"].tolist()

            # Create dataset for this batch.
            group_dataset = create_segformer_segmentation_dataset(
                batch_df, transforms=val_transforms
            )

            # Predict on the batch.
            out = trainer.predict(group_dataset)
            preds = out[0]
            true = out[1]

            # Reshape the prediction logits to tile size.
            preds = torch.from_numpy(preds).cpu()

            preds = nn.functional.interpolate(
                preds,
                size=(tile_size, tile_size),
                mode="bilinear",
                align_corners=False,
            ).argmax(
                dim=1
            )  # logits -> class predictions

            preds = preds.detach().numpy()

            # Append to the tile list.
            pred_tile_list.extend(
                [(pred, x, y) for x, y, pred in zip(x_list, y_list, preds)]
            )

            true_tile_list.extend(
                [(mask, x, y) for x, y, mask in zip(x_list, y_list, true)]
            )

        # Merge the tile masks.
        pred_gdf = merge_tile_masks(
            pred_tile_list, background_label=background_label
        )
        true_gdf = merge_tile_masks(
            true_tile_list, background_label=background_label
        )

        # Get the unique labels in either the prediction or the ground truth.
        labels = set(pred_gdf["label"].unique()) | set(
            true_gdf["label"].unique()
        )

        # Calculate the IoU for each label.
        ious = []

        for label in labels:
            pred_label_gdf = pred_gdf[pred_gdf["label"] == label]
            true_label_gdf = true_gdf[true_gdf["label"] == label]

            # If either the prediction or the ground truth is empty, the IoU is 0.
            if pred_label_gdf.empty or true_label_gdf.empty:
                ious.append(0)
            else:
                # Calculate the IoU.
                geom1 = pred_label_gdf.iloc[0]["geometry"]
                geom2 = true_label_gdf.iloc[0]["geometry"]

                intersection_area = geom1.intersection(geom2).area
                union = geom1.union(geom2).area

                ious.append(intersection_area / union)

    # Return the average of the ious.
    return float(sum(ious) / len(ious))


# def evaluate_semantic_segmentation_segformer_by_wsi(
#     model: nn.Module,
#     data: pd.DataFrame | str,
#     group_col: str,
#     label2id: dict[str, int],
#     batch_size: int = 16,
#     tile_size: int = 512,
#     background_label: int | None = None,
#     device: torch.device | None = None,
# ) -> float:
#     """Evaluate a segformer trainer model on a specific dataset provided
#     as a dataframe, grouped by a specific column. For a unique value of
#     the column it will predict on the images for that group and then
#     merge the predictions into a single multipolygon. The mean IoU will
#     be calculated agains the ground truth masks for that group. The
#     reported mean IoU will be the average of all the mean IoUs for each
#     group.

#     Args:
#         model (nn.Module): The model to evaluate.
#         data (pd.DataFrame | str): The dataframe or path to a csv file
#             containing the data to evaluate on. The dataframe must have
#             the "fp" (filepath to image) and "mask_fp" (filepath to
#             mask) columns. The column matching the "group_col" parameter
#             will be used to group tiles into groups. The "x" and "y"
#             parameters will be used for stiching the images back to
#             their original size.
#         group_col (str): The column to group the tiles by.
#         batch_size (int): The batch size to use for prediction. Note that
#             this is not the batch size use for predictions, rather it is
#             how many masks will be stored in memory. Batch size used
#             for predictions is specified on the trainer input. Default is
#             16.
#         label2id (dict): A dictionary mapping labels to their integer
#             representations.
#         tile_size (int): The size of the tiles. Default is 512.
#         background_label (int | None): The label to use for the background
#             class, which will be ignored. If None then no label will be
#             ignored. Default is None.
#         device (torch.device | None, optional): The device to use for
#             evaluation. If None, the device will be "cuda" if available,
#             otherwise "cpu". Default is None.

#     Returns:
#         float: The mean IoU across all groups.

#     """
#     # Read the data.
#     if isinstance(data, str):
#         data = pd.read_csv(data)

#     ious = []

#     if device is None:
#         device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

#     # Move the model to the device.
#     model.to(device)
#     model.eval()  # model should be in evaluation mode

#     # Predict on the images for each group separately.
#     groups = list(data[group_col].unique())
#     for group in tqdm(groups, desc="Evaluating groups"):
#         # Get the tiles in this group.
#         group_df = data[data[group_col] == group]

#         pred_tile_list = []
#         true_tile_list = []

#         # Predict on the group in batches.
#         for i in range(0, len(group_df), batch_size):
#             # Create dataset for this batch.
#             batch_df = group_df.iloc[i : i + batch_size]
#             x_list = batch_df["x"].tolist()
#             y_list = batch_df["y"].tolist()

#             # Create dataset for this batch.
#             dataset = create_segformer_segmentation_dataset(
#                 batch_df, transforms=val_transforms
#             )

#             # Get the inputs.
#             inputs = np.array(dataset[:]["pixel_values"])
#             inputs = torch.tensor(inputs, dtype=torch.float32)
#             inputs = inputs.to(device)

#             true = np.array(dataset[:]["labels"])

#             # Predict on the batch.
#             with torch.no_grad():
#                 out = model(inputs)
#                 preds = out.logits

#                 # Reshape the prediction logits to tile size.
#                 preds = torch.from_numpy(preds)

#                 preds = nn.functional.interpolate(
#                     preds,
#                     size=(tile_size, tile_size),
#                     mode="bilinear",
#                     align_corners=False,
#                 ).argmax(
#                     dim=1
#                 )  # logits -> class predictions

#                 preds = preds.detach().cpu().numpy()

#                 # Append to the tile list.
#                 pred_tile_list.extend(
#                     [(pred, x, y) for x, y, pred in zip(x_list, y_list, preds)]
#                 )

#                 true_tile_list.extend(
#                     [(mask, x, y) for x, y, mask in zip(x_list, y_list, true)]
#                 )

#         # Merge the tile masks.
#         pred_gdf = merge_tile_masks(
#             pred_tile_list, background_label=background_label
#         )
#         true_gdf = merge_tile_masks(
#             true_tile_list, background_label=background_label
#         )

#         # Calculate the Dice for each label.
#         metrics = {label: [] for label in label2id}

#         for label, integer in label2id.items():
#             if integer == background_label:
#                 continue  # don't calculate dice for background

#             # For this integer, get the prediction and ground truth.
#             pred_label_gdf = pred_gdf[pred_gdf["label"] == integer]
#             true_label_gdf = true_gdf[true_gdf["label"] == integer]

#             # If both are empty, the dice is 1.
#             if pred_label_gdf.empty and true_label_gdf.empty:
#                 metrics[label].append(1)
#             else:
#                 # Calculate the intersection and denominator.
#                 intersection = pred_label_gdf.geometry.intersection(
#                     true_label_gdf.geometry
#                 ).area.sum()

#                 # The denominator is the sum area of the two polygons.
#                 denominator = (
#                     pred_label_gdf.geometry.area.sum()
#                     + true_label_gdf.geometry.area.sum()
#                 )

#                 # The dice is 2 * intersection / denominator.
#                 metrics[label].append(float(2 * intersection / denominator))

#     # Return the dice as means and standard deviations.
#     metrics = {
#         label: {
#             "mean": float(np.mean(metrics[label])),
#             "std": float(np.std(metrics[label])),
#         }
#         for label in label2id
#     }
#     return metrics



import pandas as pd
from transformers import (
    SegformerForSemanticSegmentation,
    TrainingArguments,
    Trainer,
)
import numpy as np
from pathlib import Path

import torch
from torch import nn

from .utils import create_segformer_segmentation_dataset
from .transforms import get_train_transforms, val_transforms
from .evaluate import per_class_dice_on_dataset
from ..callbacks import MetricsLoggerCallback


def train(
    save_dir: str,
    train_data: pd.DataFrame | str,
    val_data: pd.DataFrame | str,
    label2id: dict[str, int],
    device: torch.device | None = None,
    test_data: pd.DataFrame | str | None = None,
    model_checkpoint: str | None = None,
    learning_rate: float = 1e-4,
    epochs: int = 10,
    batch_size: int = 16,
    eval_accumulation_steps: int = 100,
    max_val_size: int | None = 500,
    random_state: int = 42,
    tile_size: int = 512,
    tqdm_notebook: bool = False,
    square_symmetry_prob: float | None = 1.0,
    rotate_limit: int = 45,
    rotate_prob: float | None = 0.5,
    fill: tuple[float, ...] | float = 255.0,
    fill_mask: tuple[float, ...] | float = 0.0,
    brightness: float = 0.25,
    contrast: float = 0.25,
    saturation: float = 0.25,
    hue: float = 0.1,
):
    """Workflow for training a SegFormer semantic segmentation model.

    Args:
        save_dir (str): Directory where model checkpoint directory will
            be saved to.
        train_csv_fp (str): Train data, should be a pandas
            dataframe or a path to a csv file containing the train data.
            Must have the "fp" and "mask_fp" columns, which are the
            filepaths to the images and masks, respectively.
        val_data (pd.DataFrame | str): See train_data, but for the
            validation data.
        label2id (dict): A dictionary mapping the label names to their
            corresponding integer ids.
        device (torch.device | None, optional): Device to run the model
            on. If None, will use "cuda" if available, otherwise "cpu".
            Defaults to None.
        test_data (pd.DataFrame | str | None, optional): See train_data,
            but for the test data. Defaults to None.
        model_checkpoint (str | None, optional): A local path to a model
            checkpoint to load the model from. If None it will use the
            default weights from huggingface: "nvidia/mit-b0". Defaults
            to None.
        learning_rate (float, optional): Learning rate for the model.
            Defaults to 1e-4.
        epochs (int, optional): Number of epochs to train the model.
            Defaults to 10.
        batch_size (int, optional): Batch size for the model.
            Defaults to 16.
        eval_accumulation_steps (int, optional): Number of steps to
            accumulate the validation metrics over. Defaults to 100.
        max_val_size (int | None, optional): Maximum number of
            validation samples to use. If None, all validation samples
            will be used. Defaults to None. Note that there is a bug
            that this value must be kept short otherwise a cuda memory
            error will occur.
        random_state (int, optional): Random state for the validation
            split. Defaults to 42.
        tile_size (int, optional): Size of the tiles to use for the model.
            Defaults to 512.
        tqdm_notebook (bool, optional): Whether to use a tqdm notebook.
            Defaults to False.
        square_symmetry_prob (float | None, optional): Probability of
            applying square symmetry to the images. Defaults to 1.0.
        rotate_limit (int, optional): Maximum number of degrees to
            rotate the images. Defaults to 45.
        rotate_prob (float | None, optional): Probability of applying
            rotation to the images. Defaults to 0.5.
        fill (tuple[float, ...] | float, optional): Value to fill the
            image with. Defaults to 255.0.
        fill_mask (tuple[float, ...] | float, optional): Value to fill
            the mask with. Defaults to 0.0.
        brightness (float, optional): Brightness factor to apply to the
            images. Defaults to 0.25.
        contrast (float, optional): Contrast factor to apply to the
            images. Defaults to 0.25.
        saturation (float, optional): Saturation factor to apply to the
            images. Defaults to 0.25.
        hue (float, optional): Hue factor to apply to the images.
            Defaults to 0.1.

    Returns:
        tuple: A tuple containing the trainer and the results dictionary.

    """
    save_dir_path = Path(save_dir)

    if save_dir_path.is_dir():
        raise FileExistsError(f"Save directory {save_dir} already exists.")

    # Read the dataframes if they are strings.
    if isinstance(train_data, str):
        train_data = pd.read_csv(train_data)
    if isinstance(val_data, str):
        val_data = pd.read_csv(val_data)
    if test_data is not None and isinstance(test_data, str):
        test_data = pd.read_csv(test_data)

    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Load the default pretrained model.
    label2id = {k: int(v) for k, v in label2id.items()}
    id2label = {int(v): k for k, v in label2id.items()}

    if model_checkpoint is None:
        model = SegformerForSemanticSegmentation.from_pretrained(
            "nvidia/mit-b0", id2label=id2label, label2id=label2id
        ).to(device)
    else:
        model = SegformerForSemanticSegmentation.from_pretrained(
            model_checkpoint,
            id2label=id2label,
            label2id=label2id,
            local_files_only=True,
            ignore_mismatched_sizes=True,
        ).to(device)

    # Training arguments.
    training_args = TrainingArguments(
        str(save_dir),
        learning_rate=learning_rate,
        num_train_epochs=epochs,
        per_device_train_batch_size=batch_size,
        per_device_eval_batch_size=batch_size,
        eval_accumulation_steps=eval_accumulation_steps,
        eval_strategy="epoch",
        save_strategy="epoch",
        logging_strategy="epoch",
        load_best_model_at_end=True,
        save_total_limit=1,
    )

    # Create datasets for the model.
    train_transforms = get_train_transforms(
        square_symmetry_prob=square_symmetry_prob,
        rotate_limit=rotate_limit,
        rotate_prob=rotate_prob,
        fill=fill,
        fill_mask=fill_mask,
        brightness=brightness,
        contrast=contrast,
        saturation=saturation,
        hue=hue,
    )
    train_dataset = create_segformer_segmentation_dataset(
        train_data, transforms=train_transforms
    )

    if val_data is not None and max_val_size < len(val_data):
        val_dataset = create_segformer_segmentation_dataset(
            val_data.sample(n=max_val_size, random_state=random_state),
            transforms=val_transforms,
        )
    else:
        val_dataset = create_segformer_segmentation_dataset(
            val_data, transforms=val_transforms
        )

    def compute_metrics(eval_pred):
        """Function for computing DICE coefficient metrics for each
        class, as well as the mean DICE and weighted mean DICE (maybe).
        """
        logits, labels = eval_pred

        # Convert logits to tensor.
        logits = torch.from_numpy(logits)

        # Scale the logits back to the shape of the labels.
        logits = nn.functional.interpolate(
            logits,
            size=tile_size,
            mode="bilinear",
            align_corners=False,
        )

        # Turn the logits to predictions.
        pred = logits.argmax(dim=1)

        # Calculate the Dice for each class.
        intersection = {label: 0 for label in label2id}
        denominator = {label: 0 for label in label2id}

        # For easy calculations, convert pred to a numpy.
        pred = pred.numpy()

        # Flatten both arrays.
        pred = pred.flatten()
        labels = labels.flatten()

        # Loop through each index.
        for label, integer in label2id.items():
            pred_mask = pred == integer
            gt_mask = labels == integer

            # Calculate the interesection.
            intersection[label] += np.sum(gt_mask & pred_mask)
            denominator[label] += np.sum(gt_mask) + np.sum(pred_mask)

        # Calculate the Dice for each label.
        metrics = {}
        for label in label2id:
            denominator_value = denominator[label]

            if denominator_value:
                metrics[f"{label}_dice"] = float(
                    2 * intersection[label] / denominator_value
                )
            else:
                metrics[f"{label}_dice"] = 1

        # Add the mean Dice.
        metrics["mean_dice"] = float(np.mean(list(metrics.values())))
        return metrics

    # Trainer
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_dataset,
        eval_dataset=val_dataset,
        compute_metrics=compute_metrics,
        callbacks=[MetricsLoggerCallback],  # Add the custom callback here
    )

    # Train the model.
    _ = trainer.train()

    # Rename the checkpoint to not include a step number.
    save_dir = Path(save_dir)
    for p in save_dir.iterdir():
        if p.is_dir() and p.name.startswith("checkpoint-"):
            p.rename(save_dir / "checkpoint")

    results = {"train": {}, "val": {}}

    for log in trainer.state.log_history[:-1]:
        epoch = log["epoch"]

        if epoch % 1:
            continue  # only track end of each epoch

        # Check if this is tracking train or validation.
        if "eval_loss" in log:
            data = results["val"]
        else:
            data = results["train"]

        for k, v in log.items():
            if k.startswith("eval_"):
                k = k[5:]

            if k not in data:
                data[k] = []

            data[k].append(v)

    # For each of the datasets provided, get the metrics.
    model = trainer.model

    print("Calculating train metrics...")
    results["train"]["metrics"] = per_class_dice_on_dataset(
        model,
        train_data,
        label2id,
        batch_size=batch_size,
        device=device,
        tile_size=tile_size,
        tqdm_notebook=tqdm_notebook,
    )

    print("Calculating val metrics...")
    results["val"]["metrics"] = per_class_dice_on_dataset(
        model,
        val_data,
        label2id,
        batch_size=batch_size,
        device=device,
        tile_size=tile_size,
        tqdm_notebook=tqdm_notebook,
    )

    if test_data is not None:
        print("Calculating test metrics...")
        results["test"] = {}

        results["test"]["metrics"] = per_class_dice_on_dataset(
            model,
            test_data,
            label2id,
            batch_size=batch_size,
            device=device,
            tile_size=tile_size,
            tqdm_notebook=tqdm_notebook,
        )

    return trainer, results



"""Utility functions for DSA Helpers.

This module provides various miscellaneous utility functions that are
not grouped into their own modules.
"""

from shapely.geometry import Polygon
import numpy as np
import cv2 as cv


def remove_small_holes(
    polygon: Polygon, hole_area_threshold: float
) -> Polygon:
    """Remove small holes from a shapely polygon.

    Args:
        polygon (shapely.geometry.Polygon): Polygon to remove holes
            from.
        hole_area_threshold (float): Minimum area of a hole to keep it.

    Returns:
        shapely.geometry.Polygon: Polygon with small holes removed.

    """
    if not polygon.interiors:  # if there are no holes, return as is
        return polygon

    # Filter out small holes
    new_holes = [
        hole
        for hole in polygon.interiors
        if Polygon(hole).area > hole_area_threshold
    ]

    # Create a new polygon with only large holes
    return Polygon(polygon.exterior, new_holes)


def convert_to_json_serializable(
    data: int | float | str | list | dict,
) -> int | float | str | list | dict:
    """Convert a list, integer, float, or dictionary into a JSON
    serializable version of the object. Uses recursion to make sure the
    entire input data structure is converted to a JSON serializable
    version.

    Args:
        data (int | float | str | list | dict): Data to convert to
            JSON serializable.

    Returns:
        int | float | str | list | dict: JSON serializable version of
            the input data.

    """
    if isinstance(data, np.ndarray):
        return data.tolist()
    elif isinstance(data, np.integer):
        return int(data)
    elif isinstance(data, np.floating):
        return float(data)
    elif isinstance(data, dict):
        return {
            key: convert_to_json_serializable(value)
            for key, value in data.items()
        }
    elif isinstance(data, list):
        return [convert_to_json_serializable(item) for item in data]
    return data


def contours_to_polygons(binary_image: np.ndarray) -> list[Polygon]:
    """Convert a binary image to a list of polygons.

    Args:
        binary_image (numpy.ndarray): Binary image.

    Returns:
        list[shapely.geometry.Polygon]: List of polygons.

    """
    contours, hierarchy = cv.findContours(
        binary_image, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE
    )
    hierarchy = hierarchy[0]  # shape: (n_contours, 4)

    def contour_to_coords(contour):
        return [(int(p[0][0]), int(p[0][1])) for p in contour]

    polygons = []
    used = set()

    for i, (next_idx, prev_idx, child_idx, parent_idx) in enumerate(hierarchy):
        if parent_idx != -1:
            # Skip child contours for now (they'll be added as holes)
            continue

        # Outer contour
        exterior = contour_to_coords(contours[i])

        if len(exterior) < 4:
            # Skip contours that are too small
            used.add(i)
            continue

        holes = []

        # Look for children (holes)
        child = hierarchy[i][2]
        while child != -1:
            hole_coords = contour_to_coords(contours[child])
            holes.append(hole_coords)
            used.add(child)
            # Check for nested children (e.g. box in a hole in a box)
            grandchild = hierarchy[child][2]
            if grandchild != -1:
                # Treat grandchild as a new polygon later
                pass
            child = hierarchy[child][0]  # next sibling hole

        used.add(i)
        polygon = Polygon(exterior, holes)

        if not polygon.is_valid:
            # Attempt to fix the polygon
            polygon = polygon.buffer(0)

        if polygon.is_valid:
            polygons.append(polygon)

    # Add any unused outer contours (e.g. nested objects)
    for i in range(len(contours)):
        if i not in used:
            coords = contour_to_coords(contours[i])

            if len(coords) < 4:
                # Skip contours that are too small
                continue

            poly = Polygon(coords)

            if not poly.is_valid:
                # Attempt to fix the polygon
                poly = poly.buffer(0)

            if poly.is_valid:
                polygons.append(poly)

    return polygons



import numpy as np
from shapely.geometry import Polygon


def convert_box_type(box):
    """Convert a box type from YOLO format (x-center, y-center, box-width, box-height) to (x1, y1, x2, y2) where point 1 is the
    top left corner of box and point 2 is the bottom right corner

    INPUT
    -----
    box : array
        [N, 4], each row a point and the format being (x-center, y-center, box-width, box-height)

    RETURN
    ------
    new_box : array
        [N, 4] each row a point and the format x1, y1, x2, y2

    """
    # get half the box height and width
    half_bw = box[:, 2] / 2
    half_bh = box[:, 3] / 2

    new_box = np.zeros(box.shape, dtype=box.dtype)
    new_box[:, 0] = box[:, 0] - half_bw
    new_box[:, 1] = box[:, 1] - half_bh
    new_box[:, 2] = box[:, 0] + half_bw
    new_box[:, 3] = box[:, 1] + half_bh

    return new_box


def read_yolo_label(filepath, im_shape=None, shift=None, convert=False):
    """Read a yolo label text file. It may contain a confidence value for the labels or not, will handle both cases

    INPUTS
    ------
    filepath : str
        the path of the text file
    im_shape : tuple or int (default: None)
        image width and height corresponding to the label, if an int it is assumed both are the same. Will scale coordinates
        to int values instead of normalized if given
    shift : tuple or int (default: None)
        shift value in the x and y direction, if int it is assumed to be the same in both. These values will be subtracted and applied
        after scaling if needed
    convert : bool (default: False)
        If True, convert the output boxes from yolo format (label, x-center, y-center, width, height, conf) to (label, x1, y1, x2, y2, conf)
        where point 1 is the top left corner of box and point 2 is the bottom corner of box

    RETURN
    ------
    coords : array
        coordinates array, [N, 4 or 5] depending if confidence was in input file

    """
    coords = []

    with open(filepath, "r") as fh:
        for line in fh.readlines():
            if len(line):
                coords.append([float(ln) for ln in line.strip().split(" ")])

    coords = np.array(coords)

    # scale coords if needed
    if im_shape is not None:
        if isinstance(im_shape, int):
            w, h = im_shape, im_shape
        else:
            w, h = im_shape[:2]

        coords[:, 1] *= w
        coords[:, 3] *= w
        coords[:, 2] *= h
        coords[:, 4] *= h

    # shift coords
    if shift is not None:
        if isinstance(shift, int):
            x_shift, y_shift = shift, shift
        else:
            x_shift, y_shift = shift[:2]

        coords[:, 1] -= x_shift
        coords[:, 2] -= y_shift

    if convert:
        coords[:, 1:5] = convert_box_type(coords[:, 1:5])

    return coords


def corners_to_polygon(x1: int, y1: int, x2: int, y2: int) -> Polygon:
    """Return a Polygon from shapely with the box coordinates given the top left and bottom right corners of a
    rectangle (can be rotated).

    Args:
        x1, y1, x2, y2: Coordinates of the top left corner (point 1) and the bottom right corner (point 2) of a box.

    Returns:
        Shapely polygon object of the box.

    """
    return Polygon([(x1, y1), (x2, y1), (x2, y2), (x1, y2)])



# A test application.
from dash import Dash, html
import dash_bootstrap_components as dbc
from dsa_helpers.dash.header import get_header
import os

os.environ["DSA_API_URL"] = "https://megabrain.neurology.emory.edu/api/v1"

app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])

app.layout = html.Div([get_header("Test")])

if __name__ == "__main__":
    app.run_server(debug=True)



"""Image utility functions for DSA Helpers.

This module provides various utility functions for image processing and
manipulation.

It includes functions for:
- Exctracting contours from binary masks, and formatting them into a
GeoDataFrame.

Note:
    This module requires external dependencies such as OpenCV and
    shapely.
"""

import numpy as np
import cv2 as cv
from .imread import imread

from shapely.affinity import translate
from shapely.geometry import Polygon


def label_mask_to_polygons(
    mask: str | np.ndarray,
    x_offset: int = 0,
    y_offset: int = 0,
    exclude_labels: int | list[int] = 0,
    min_area: int = 0,
) -> list[Polygon, int]:
    """
    Extract contours from a label mask and convert them into shapely
    polygons.

    Args:
        mask (str | np.ndarray): Path to the mask image or the mask.
        x_offset (int): Offset to add to x coordinates of polygons.
        y_offset (int): Offset to add to y coordinates of polygons.
        exclude_labels (int | list[int]): Label(s) to exclude from the
            output.
        min_area (int): Minimum area of polygons to include.

    Returns:
        list[Polygon, int]: List of polygons and their corresponding
            labels.

    """
    # Convert exclude_labels to a list if it's an integer.
    if isinstance(exclude_labels, int):
        exclude_labels = [exclude_labels]

    # Read the mask if it's a path.
    if isinstance(mask, str):
        mask = imread(mask, grayscale=True)

    # Unique labels, excluding those specified in exclude_labels.
    labels = [
        label for label in np.unique(mask) if label not in exclude_labels
    ]

    polygons = []  # track all polygons

    # Loop through unique label index.
    for label in labels:
        # Filter to mask for this label.
        label_mask = (mask == label).astype(np.uint8)

        # Find contours.
        contours, hierarchy = cv.findContours(
            label_mask, cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE
        )

        # Process the contours.
        polygons_dict = {}

        for idx, (contour, h) in enumerate(zip(contours, hierarchy[0])):
            # The contour should contain at least three points, otherwise it's
            # not a polygon.
            if len(contour) > 3:
                # To include holes, group the polygons by their parent index.
                if idx not in polygons_dict:
                    polygons_dict[idx] = {"holes": []}

                if h[3] == -1:
                    # This does not have a parent so its a polygon.
                    polygons_dict[idx]["polygon"] = contour.reshape(-1, 2)
                else:
                    # This has a parent so its a hole.
                    polygons_dict[h[3]]["holes"].append(contour.reshape(-1, 2))

        # Convert to shapely polygon objects.
        for data in polygons_dict.values():
            if "polygon" in data:
                polygon = Polygon(data["polygon"], holes=data["holes"])

                # Shift the polygon by the offset.
                polygon = translate(polygon, xoff=x_offset, yoff=y_offset)

                # Include the polygon if it's greater than the minimum area.
                if polygon.area >= min_area:
                    polygons.append([polygon, int(label)])

    return polygons


def convert_label_mask_to_rgb(
    mask: np.ndarray,
    int2rgb: dict,
    alpha: bool = False,
    fill_color: tuple[int, int, int] | tuple[int, int, int, int] = (
        255,
        255,
        255,
    ),
) -> np.ndarray:
    """
    Convert a label mask to an RGB mask.

    Args:
        mask: Label mask to convert.
        int2rgb: Dictionary mapping integer labels to RGB colors. If
            passing alpha as True, then pass RGBA colors.
        alpha: Whether to include the alpha channel in the RGB mask.
        fill_color: Color to fill the mask with.

    Returns:
        (np.ndarray): RGB(A) mask.

    """
    # Create a new RGB mask of the same size as the label mask.
    rgb_mask = np.zeros(
        (mask.shape[0], mask.shape[1], 4 if alpha else 3), dtype=np.uint8
    )

    rgb_mask[:] = fill_color

    # Loop through the label mask.
    for label, rgb in int2rgb.items():
        rgb_mask[mask == label] = rgb

    return rgb_mask



import torch
from large_image_eager_iterator import LargeImagePrefetch
from transformers import (
    SegformerForSemanticSegmentation,
    SegformerImageProcessor,
)
from PIL import Image
from large_image import getTileSource
import geopandas as gpd
from shapely.affinity import scale
import pandas as pd

from ...image_utils import label_mask_to_polygons
from ...gpd_utils import remove_gdf_overlaps


def inference(
    model: str | torch.nn.Module,
    wsi_fp: str,
    batch_size: int = 16,
    tile_size: int = 512,
    mag: float | None = None,
    workers: int = 8,
    chunk_mult: int = 2,
    prefetch: int = 2,
    device: str | None = None,
    tolerance: float | None = 2.0,
    min_area: float = 1000,
) -> gpd.GeoDataFrame:
    """Inference using SegFormer semantic segmentation model on a WSI.

    Args:
        model (str | torch.nn.Module): Path to the model checkpoint or
            a pre-loaded model.
        wsi_fp (str): File path to the WSI.
        batch_size (int, optional): Batch size for inference. Defaults
            to 16.
        tile_size (int, optional): Tile size for inference. Defaults to
            512.
        mag (float, optional): Magnification for inference. Defaults to
            None, which will use the scan magnification of WSI.
        workers (int, optional): Number of workers for inference.
            Defaults to 8.
        chunk_mult (int, optional): Chunk multiplier for inference.
            Defaults to 2.
        prefetch (int, optional): Number of prefetch for inference.
            Defaults to 2.
        device (str, optional): Device for inference. Default is None,
            will use "gpu" if available, otherwise "cpu".
        tolerance (float | None, optional): Tolerance for simplification
            of the predicted polygons. Guidelines: 0.5 for mild
            simplification, 1.0 for moderate, 2.0 for more aggressive.
            If set to None or 0.0, the polygons will not be simplified.
        min_area (float, optional): Minimum area of a polygon to be
            simplified. This is added to avoid simplifying small
            polygons that would probably lose too much information.
            Defaults to 1000.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing the predicted
            polygons and labels.

    """
    # Initiate the tile iterator.
    iterator = LargeImagePrefetch(
        wsi_fp,
        batch=batch_size,
        tile_size=(tile_size, tile_size),
        scale_mode="mag",
        target_scale=mag,
        workers=workers,
        chunk_mult=chunk_mult,
        prefetch=prefetch,
        nchw=False,
        icc=True,
    )

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
        print(f"Using device: {device}")

    device = torch.device(device)

    # Load the model.
    if isinstance(model, str):
        model = SegformerForSemanticSegmentation.from_pretrained(
            model, local_files_only=True, device_map=device
        )

    model.eval()

    # Iterate through batches.
    batch_n = 0

    # Image processor for images.
    processor = SegformerImageProcessor()

    # Track all predicted polygons.
    wsi_polygons = []

    # Scaling factor, multiply to go from scan magnification to desired mag.
    ts_metadata = getTileSource(wsi_fp).getMetadata()
    scan_mag = ts_metadata["magnification"]

    if mag is None:
        mag = scan_mag
        sf = 1.0
    else:
        sf = mag / scan_mag

    for batch in iterator:
        # Get the batch of images.
        imgs = batch[0].view()  # returns a numpy array of shape (N, H, W, C)
        coordinates = batch[1]

        # Convert the numpy arrays to PIL images.
        imgs = [Image.fromarray(img) for img in imgs]

        # Pass the images through the processor.
        inputs = processor(imgs, return_tensors="pt")
        inputs = inputs.to(model.device)

        # Predict on the batch.
        with torch.no_grad():
            output = model(inputs["pixel_values"])
            logits = output.logits

            # Get the logits out, resizing them to the original tile size.
            logits = torch.nn.functional.interpolate(
                logits,
                size=tile_size,
                mode="bilinear",
            )

            # Get predicted class labels for each pixel.
            masks = torch.argmax(logits, dim=1).detach().cpu().numpy()

        # Loop through each mask to extract the contours as shapely polygons.
        for i, mask in enumerate(masks):
            img_metadata = coordinates[i]
            x, y = img_metadata[6], img_metadata[4]
            x = int(x * sf)
            y = int(y * sf)

            polygon_and_labels = label_mask_to_polygons(
                mask,
                x_offset=x,
                y_offset=y,
            )

            for polygon_and_label in polygon_and_labels:
                polygon, label = polygon_and_label
                label = int(label)

                # Do something with the polygon and label.
                wsi_polygons.append([polygon, label])

        batch_n += 1
        print(f"\r    Processed batch {batch_n}.    ", end="")
    print()

    # Convert polygons and labels to a GeoDataFrame.
    gdf = gpd.GeoDataFrame(wsi_polygons, columns=["geometry", "label"])
    gdf["geometry"] = gdf["geometry"].buffer(1)
    gdf = gdf.dissolve(by="label", as_index=False)
    gdf = gdf.explode(index_parts=False).reset_index(drop=True)

    if tolerance:
        # Add the area column.
        gdf["area"] = gdf["geometry"].area

        # Split the dataframe into those polygons that need to be simplified and those that do not.
        gdf_to_simplify = gdf[gdf["area"] > min_area].reset_index(drop=True)
        gdf_not_to_simplify = gdf[gdf["area"] <= min_area].reset_index(
            drop=True
        )

        # Simplify the polygons that need to be simplified.
        gdf_to_simplify["geometry"] = gdf_to_simplify["geometry"].simplify(
            tolerance=tolerance, preserve_topology=True
        )

        # Concatenate the simplified polygons with the polygons that do not need to be simplified.
        gdf = pd.concat(
            [gdf_to_simplify, gdf_not_to_simplify], ignore_index=True
        )

    # Remove overlapping polygons.
    no_overlap_gdf = remove_gdf_overlaps(
        gdf,
        columns="label",
    )

    # Scale the polygon to scan magnification.
    no_overlap_gdf["geometry"] = no_overlap_gdf["geometry"].apply(
        lambda geom: scale(geom, xfact=1 / sf, yfact=1 / sf, origin=(0, 0))
    )

    return no_overlap_gdf



# Callbacks for the header component.
from dash import callback, Output, Input, State, no_update
from girder_client import GirderClient


def get_callbacks(dsa_api_url: str, store_id: str):
    """Setup the callbacks to the DSA API url of interest and the proper store id."""

    @callback(Output(f"{store_id}=login-btn", "children"), Input(store_id, "data"))
    def check_user_store(data):
        # Check if the user store has user info or if no one is logged in.
        return data["user"] if len(data) else "Log in"

    @callback(
        [
            Output(f"{store_id}=login-modal", "is_open", allow_duplicate=True),
            Output(f"{store_id}=logout-modal", "is_open", allow_duplicate=True),
        ],
        [
            Input(f"{store_id}=login-btn", "n_clicks"),
            State(f"{store_id}=login-btn", "children"),
        ],
        prevent_initial_call=True,
    )
    def open_login_modal(n_clicks, children):
        # Open login / logout modal.
        if n_clicks:
            if children == "Log in":
                return True, False
            else:
                return False, True

        return False, False

    @callback(
        [
            Output(store_id, "data"),
            Output(f"{store_id}=login-failed", "hidden", allow_duplicate=True),
            Output(f"{store_id}=login-modal", "is_open", allow_duplicate=True),
            Output(f"{store_id}=login", "value", allow_duplicate=True),
            Output(f"{store_id}=password", "value", allow_duplicate=True),
        ],
        [
            Input(f"{store_id}=log-in-btn", "n_clicks"),
            State(f"{store_id}=login", "value"),
            State(f"{store_id}=password", "value"),
        ],
        prevent_initial_call=True,
    )
    def login(n_clicks, login, password):
        # Try to login.
        gc = GirderClient(apiUrl=dsa_api_url)

        try:
            _ = gc.authenticate(username=login, password=password)

            response = gc.get("token/session")

            user = gc.get("user/me")["login"]

            return {"user": user, "token": response["token"]}, True, False, "", ""
        except:
            return (
                {},
                False,
                True,
                no_update,
                no_update,
            )

    @callback(
        [
            Output(f"{store_id}=login-modal", "is_open", allow_duplicate=True),
            Output(f"{store_id}=login", "value", allow_duplicate=True),
            Output(f"{store_id}=password", "value", allow_duplicate=True),
            Output(f"{store_id}=login-failed", "hidden", allow_duplicate=True),
        ],
        Input(f"{store_id}=close-login-modal", "n_clicks"),
        prevent_initial_call=True,
    )
    def close_login_modal(n_clicks):
        if n_clicks:
            return False, "", "", True

        return False, "", "", True

    @callback(
        [
            Output(store_id, "data", allow_duplicate=True),
            Output(f"{store_id}=logout-modal", "is_open", allow_duplicate=True),
        ],
        Input(f"{store_id}=logout-btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def logout(n_clicks):
        if n_clicks:
            return {}, False

        return no_update, False



from torch.utils.data import Dataset
import pandas as pd
from PIL import Image


class SegFormerSegmentationDataset(Dataset):
    """PyTorch dataset class for semantic segmentation using
    HuggingFaces SegFormer model.

    """

    def __init__(
        self, df: pd.DataFrame, extras: bool = False, group: str = "wsi_name"
    ):
        """Initiate an instance of the segmentation dataset.

        Args:
            df (pandas.DataFrame): Must have columns "fp" and "mask_fp".
            extras (bool, optional): Whether to include additional
                columns "x", "y" and "group" in the dataset. Default
                is False.
            group (str | None, optional): Additional column to add when
                extras is True. Default is "wsi_name".

        """
        self.image_files = df["fp"].tolist()
        self.mask_files = df["mask_fp"].tolist()
        self.extras = extras

        if extras:
            self.x = df["x"].tolist()
            self.y = df["y"].tolist()
            self.group = df[group].tolist()

    def __len__(self):
        return len(self.image_files)

    def __getitem__(self, idx):
        # Get the filepath to the image and its mask.
        image_path = self.image_files[idx]
        mask_path = self.mask_files[idx]

        if self.extras:
            return {
                "pixel_values": Image.open(image_path),
                "label": Image.open(mask_path),
                "x": self.x[idx],
                "y": self.y[idx],
                "group": self.group[idx],
            }
        else:
            return {
                "pixel_values": Image.open(image_path),
                "label": Image.open(mask_path),
            }



from transformers import SegformerImageProcessor
from torchvision.transforms import ColorJitter
import albumentations as A
from typing import Callable
import numpy as np
from PIL import Image


def val_transforms(example_batch):
    """Default transforms for validation images."""
    processor = SegformerImageProcessor()

    images = [x for x in example_batch["pixel_values"]]
    labels = [x for x in example_batch["label"]]

    inputs = processor(images, labels)

    return inputs


def train_transforms(example_batch):
    """Default transforms for training images."""

    processor = SegformerImageProcessor()
    jitter = ColorJitter(
        brightness=0.25, contrast=0.25, saturation=0.25, hue=0.1
    )

    images = [jitter(x) for x in example_batch["pixel_values"]]
    labels = [x for x in example_batch["label"]]

    inputs = processor(images, labels)

    return inputs


def get_train_transforms(
    square_symmetry_prob: float | None = 1.0,
    rotate_limit: int = 45,
    rotate_prob: float | None = 0.5,
    fill: tuple[float, ...] | float = 255.0,
    fill_mask: tuple[float, ...] | float = 0.0,
    brightness: float = 0.25,
    contrast: float = 0.25,
    saturation: float = 0.25,
    hue: float = 0.1,
) -> Callable:
    """Get a train transform function that can be used to transform a
    batch of images and labels. Used with batches for segformer
    semantic segmentation models.

    Args:
        square_symmetry_prob (float | None, optional): Probability of
            applying square symmetry. Default is 1.0.
        rotate_limit (int, optional): Maximum rotation limit in degrees.
            Default is 45.
        rotate_prob (float | None, optional): Probability of applying
            rotation. Default is 0.5.
        fill (tuple[float, ...] | float, optional): Fill value for the
            image. Default is 255.0.
        fill_mask (tuple[float, ...] | float, optional): Fill value for
            the mask. Default is 0.0.
        brightness (float, optional): Brightness adjustment factor.
            Default is 0.25.
        contrast (float, optional): Contrast adjustment factor.
            Default is 0.25.
        saturation (float, optional): Saturation adjustment factor.
            Default is 0.25.
        hue (float, optional): Hue adjustment factor. Default is 0.1.

    Returns:
        Callable: A function that can be used to transform a batch of
            images and labels.

    """
    albumentation_pipeline = []

    if square_symmetry_prob is not None:
        albumentation_pipeline.append(A.SquareSymmetry(p=square_symmetry_prob))

    if rotate_prob is not None:
        albumentation_pipeline.append(
            A.Rotate(
                limit=rotate_limit,
                fill=fill,
                p=rotate_prob,
                fill_mask=fill_mask,
            )
        )

    if len(albumentation_pipeline):
        albumentation_pipeline = A.Compose(albumentation_pipeline)
    else:
        albumentation_pipeline = None

    jitter = ColorJitter(
        brightness=brightness,
        contrast=contrast,
        saturation=saturation,
        hue=hue,
    )

    processor = SegformerImageProcessor()

    def transform(batch):
        if albumentation_pipeline is not None:
            images, labels = [], []

            # Pass through the albumentation pipeline.
            for img, label in zip(batch["pixel_values"], batch["label"]):
                # Apply the albumentation pipeline.
                img = np.array(img)
                label = np.array(label)

                augmented = albumentation_pipeline(image=img, mask=label)
                images.append(Image.fromarray(augmented["image"]))
                labels.append(Image.fromarray(augmented["mask"]))
        else:
            images = batch["pixel_values"]
            labels = batch["label"]

        # Apply color jitter.
        images = [jitter(x) for x in images]

        # Process the images and labels.
        return processor(images, labels)

    return transform
